SystemOrganization addCategory: #'Collections-BitmapCharacterSet'!

Collection subclass: #BitmapCharacterSet
	instanceVariableNames: 'byteCharacters wideCharacters tally'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-BitmapCharacterSet'!

!BitmapCharacterSet commentStamp: 'monty 12/10/2014 12:59' prior: 0!
This class implements a set of Character objects similar to CharacterSet and WideCharacterSet, but it uses a bitmap internally to test if wide (multibyte) characters belong to it rather than using a Dictionary like WideCharacterSet does. For byte characters, a simple 256-element Array is used, the same as with CharacterSet, which is faster but uses more memory.!

!BitmapCharacterSet class methodsFor: 'as yet unclassified' stamp: 'monty 10/29/2014 08:11'!
inspectorClass
	^ Set inspectorClass! !

!BitmapCharacterSet class methodsFor: 'instance creation' stamp: 'JAAyer 1/19/2011 23:03'!
new
	^ self new: 256! !

!BitmapCharacterSet class methodsFor: 'instance creation' stamp: 'JAAyer 1/19/2011 23:04'!
new: aCapacity
	^ self basicNew initialize: aCapacity! !

!BitmapCharacterSet class methodsFor: 'instance creation' stamp: 'JAAyer 1/20/2011 00:52'!
newFrom: aCollection
	^ self new
		addAll: aCollection;
		yourself	! !

!BitmapCharacterSet methodsFor: 'comparing' stamp: 'JAA 3/31/2011 17:17'!
= anObject
	self == anObject
		ifTrue: [^ true].

	(self class == anObject class
		and: [tally = anObject size])
		ifFalse: [^ false].

	^ self allSatisfy: [:each | anObject includes: each].! !

!BitmapCharacterSet methodsFor: 'adding' stamp: 'monty 12/10/2014 12:50'!
add: aCharacter
	| asciiValue |

	(asciiValue := aCharacter asciiValue) < 256
		ifTrue: [
			(byteCharacters at: asciiValue + 1)
				ifFalse: [tally := tally + 1].
			byteCharacters at: asciiValue + 1 put: true]
		ifFalse: [| byteIndex byte bitmask |
			byteIndex := (asciiValue // 8) - 32 + 1.
			(wideCharacters isNil or: [byteIndex > wideCharacters size])
				ifTrue: [self growWideCharacterBitmapTo: (byteIndex * 1.5) asInteger].

			byte := wideCharacters at: byteIndex.
			bitmask := 16r80 bitShift: (asciiValue \\ 8) negated.
			
			"Only increment the tally if the character is not already present"
			(byte bitAnd: bitmask) > 0
				ifFalse: [tally := tally + 1].

			wideCharacters at: byteIndex put: (byte bitOr: bitmask)].
	^ aCharacter.! !

!BitmapCharacterSet methodsFor: 'accessing' stamp: 'Anonymous 11/12/2013 00:32'!
capacity
	^ byteCharacters size +
		(wideCharacters
			ifNil: [0]
			ifNotNil: [wideCharacters size * 8]).! !

!BitmapCharacterSet methodsFor: 'enumerating' stamp: 'monty 12/10/2014 13:02'!
do: aBlock
	1 to: byteCharacters size do: [:i |
		(byteCharacters at: i)
			ifTrue: [aBlock value: (Character value: i - 1)]].

	wideCharacters
		ifNil: [^ self].
	1 to: wideCharacters size do: [:byteIndex |
		0 to: 7 do: [:shiftIndex |
			((wideCharacters at: byteIndex) bitAnd:
				(16r80 bitShift: shiftIndex negated)) > 0
				ifTrue: [
					aBlock value:
						(Character value:
							(byteIndex - 1) * 8 + 256 + shiftIndex)]]].! !

!BitmapCharacterSet methodsFor: 'private' stamp: 'Anonymous 11/12/2013 00:48'!
growWideCharacterBitmapTo: aSizeInBytes
	| newSize |

	newSize := aSizeInBytes min: (self lastCodePoint // 8) - 32 + 1.
	wideCharacters
		ifNil: [
			wideCharacters := ByteArray new: newSize.
			^ self].

	wideCharacters :=
		(ByteArray new: newSize)
			replaceFrom: 1
			to: wideCharacters size
			with: wideCharacters
			startingAt: 1.! !

!BitmapCharacterSet methodsFor: 'testing' stamp: 'JAAyer 1/20/2011 01:09'!
hash
	| hash |

	hash := byteCharacters hash.
	wideCharacters
		ifNotNil: [hash := hash bitXor: wideCharacters hash].
	^ hash.! !

!BitmapCharacterSet methodsFor: 'testing' stamp: 'monty 12/10/2014 12:52'!
includes: aCharacter
	| asciiValue |

	(asciiValue := aCharacter asciiValue) < 256
		ifTrue: [^ (byteCharacters at: asciiValue + 1)]
		ifFalse: [| byteIndex bitmask |
			byteIndex := (asciiValue // 8) - 32 + 1.
			(wideCharacters == nil
				or: [byteIndex > wideCharacters size])
				ifTrue: [^ false].
			bitmask := 16r80 bitShift: (asciiValue \\ 8) negated.
			^ ((wideCharacters at: byteIndex) bitAnd: bitmask) > 0]! !

!BitmapCharacterSet methodsFor: 'initialization' stamp: 'monty 12/10/2014 12:50'!
initialize: aCapacity
	byteCharacters := Array new: 256 withAll: false.
	aCapacity > 256
		ifTrue: [self growWideCharacterBitmapTo: ((aCapacity - 1) // 8) - 32 + 1].
	tally := 0.! !

!BitmapCharacterSet methodsFor: 'private' stamp: 'Anonymous 11/12/2013 00:31'!
lastCodePoint
	^ 16r10FFFF! !

!BitmapCharacterSet methodsFor: 'copying' stamp: 'JAAyer 1/20/2011 01:05'!
postCopy
	byteCharacters := byteCharacters copy.
	wideCharacters
		ifNotNil: [wideCharacters := wideCharacters copy].! !

!BitmapCharacterSet methodsFor: 'removing' stamp: 'monty 12/10/2014 12:52'!
remove: aCharacter ifAbsent: aBlock
	| asciiValue |

	(asciiValue := aCharacter asciiValue) < 256
		ifTrue: [
			(byteCharacters at: asciiValue + 1)
				ifFalse: [^ aBlock value].
			byteCharacters at: asciiValue + 1 put: false]
		ifFalse: [| byteIndex byte bitmask |
			byteIndex := (asciiValue // 8) - 32 + 1.
			(wideCharacters isNil or: [byteIndex > wideCharacters size])
				ifTrue: [^ aBlock value].

			bitmask := 16r80 bitShift: (asciiValue \\ 8) negated.
			((byte := wideCharacters at: byteIndex) bitAnd: bitmask) > 0
				ifFalse: [^ aBlock value].

			wideCharacters at: byteIndex put: (byte bitAnd: bitmask bitInvert)].
	tally := tally - 1.
	^ aCharacter.! !

!BitmapCharacterSet methodsFor: 'accessing' stamp: 'JAAyer 1/21/2011 17:29'!
size
	^ tally! !
